/*
Copyright (c) 2006, A Small Orange Software
 All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

- Neither the name of A Small Orange Software nor the names of its contributors may
  be used to endorse or promote products derived from this software without
  specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
THE POSSIBILITY OF SUCH DAMAGE.
*/

#define _GNU_SOURCE
#include <sys/types.h>

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <stdio.h>

#include "server.h"
#include "connections.h"
#include "response.h"
#include "connections.h"
#include "log.h"

#include "plugin.h"

typedef struct {
	buffer *status_url;
	buffer *state_file;

	buffer *limit_buf;
	buffer *period_buf;

	off_t limit;
	off_t period;
	
	time_t start_ts, *start_ts_ptr;
	
	off_t requests, *requests_ptr;
	off_t kbytes_written, *kbytes_written_ptr; /* Store in KB's to keep 32 bit systems from crying */
	off_t written_overflow, *written_overflow_ptr; /* An overflow counter for sub-1K requests and for accuracy */
	off_t kbytes_read, *kbytes_read_ptr; /* We'll track this for now (since it's easy cheesy), but only use the outgoing data for limiting */
} plugin_config;

typedef struct {
	PLUGIN_DATA;
	
	plugin_config **config_storage;
	plugin_config conf; 
} plugin_data;

INIT_FUNC(mod_bandwidth_init) {
	plugin_data *p;
	
	p = calloc(1, sizeof(*p));
	
	return p;
}

FREE_FUNC(mod_bandwidth_free) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;
	
	
	if (p->config_storage) {
		size_t i;
		for (i = 0; i < srv->config_context->used; i++) {
			plugin_config *s = p->config_storage[i];
			
			buffer_free(s->status_url);
			buffer_free(s->state_file);
			buffer_free(s->limit_buf);
			buffer_free(s->period_buf);
			
			free(s);
		}
		free(p->config_storage);
	}
	
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_bandwidth_set_defaults) {
	plugin_data *p = p_d;
	size_t i;
	FILE *fp;
	char *ctx_name;
	int rc = 0;
	
	config_values_t cv[] = { 
		{ "bw.status-url", NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },
		{ "bw.state-file", NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },
		{ "bw.limit",      NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ "bw.period",     NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ NULL,            NULL, T_CONFIG_UNSET,  T_CONFIG_SCOPE_UNSET }
	};
	
	if (!p) return HANDLER_ERROR;
	
	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
	
	for (i = 0; i < srv->config_context->used; i++) {
	    plugin_config *s;
		s = calloc(1, sizeof(plugin_config));
		s->status_url = buffer_init();
		s->state_file = buffer_init();
		
		s->limit_buf = buffer_init();
		s->limit = 0;
		s->period_buf = buffer_init();
		s->period = 0;
		
		s->start_ts = srv->cur_ts;
		
		s->requests = 0;
		s->kbytes_written = 0;
		s->written_overflow = 0;
		s->kbytes_read = 0;
		
		cv[0].destination = s->status_url;
		cv[1].destination = s->state_file;
		cv[2].destination = s->limit_buf;
		cv[3].destination = s->period_buf;
		
		p->config_storage[i] = s;
	
		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
			return HANDLER_ERROR;
		}
		
		/* Check global options */
		if (i > 0 && !buffer_is_empty(s->status_url)) {
			log_error_write(srv, __FILE__, __LINE__, "s", 
			          		"bw.status_url can only be set as a global option.");
			
			return HANDLER_ERROR;
		}
		
		if (i > 0 && !buffer_is_empty(s->state_file)) {
			log_error_write(srv, __FILE__, __LINE__, "s", 
			           		"bw.state_file can only be set as a global option.");
			
			return HANDLER_ERROR;
		}
	}
	
	p->conf.state_file = p->config_storage[0]->state_file;
	
	/* State_file is required */
	if (buffer_is_empty(p->conf.state_file)) {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"bw.state_file has to be set");
		return HANDLER_ERROR;
	}
	
	/* Load in any saved state so we can restart without losing the used bandwidth for a context */
	if ((fp = fopen(p->conf.state_file->ptr, "a+")) == NULL) {
	    log_error_write(srv, __FILE__, __LINE__, "sb"
	                    "Cannot open state file:", p->conf.state_file);
	    return HANDLER_ERROR;
	}
	
	ctx_name = calloc(1000,sizeof(char));

    /* Read each line and store the saved values into the context */
	while (fscanf(fp, "%s ", ctx_name) == 1) {
	    plugin_config *s = NULL;
	    
        log_error_write(srv, __FILE__, __LINE__, "ss", "Reading state file line:", ctx_name);
        
        /* Find the context this line is for */
        for (i = 0; i < srv->config_context->used; i++) {
            data_unset *du = srv->config_context->data[i];
            s = p->config_storage[i];
            if(buffer_is_equal_string(du->key, ctx_name, strlen(ctx_name)))
                break;
            s = NULL;
        }

        /* If we don't have this context anymore, skip it */
        if(s == NULL) {
         	(void) fscanf(fp, "%*[^\n]");
         	log_error_write(srv, __FILE__, __LINE__, "ss", "State: Removed context",ctx_name);
			continue;
		}
        
        /* Now actually read in the values */
	    rc = fscanf(fp, "%ld %ld %ld %ld\n",
	                &(s->start_ts), 
	                &(s->kbytes_written), 
	                &(s->written_overflow), 
	                &(s->requests));              
	              
	    if( rc != 4 )
	        break;
	}
	
	fclose(fp);
	
	return HANDLER_GO_ON;
}

static handler_t mod_bandwidth_handle_status(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	buffer *b;
	time_t ts;
	int days, hours, mins, seconds;
	
	b = chunkqueue_get_append_buffer(con->write_queue);

	BUFFER_COPY_STRING_CONST(b, 
                 "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
                 "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
                 "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
                 "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
                 " <head>\n");
	
	BUFFER_APPEND_STRING_CONST(b,
                 "<title>Bandwidth Usage</title>\n"
                 "<style type='text/css'>\n"
				 "body{ font-family: Arial, sans-serif; }\n"
                 "div.foot { \n"
                 "    font-family: 'Courier New', Courier, monospace; font-size: 10pt; color: #787878;\n"
                 "    border-top: 1px solid #646464; margin-top: 10px; padding-top: 5px;\n}\n"
                 "table { border: solid 1px black; }\n"
                 "td { padding: 3px; }\n"
                 "td.head { text-align: center; background: black; color: white; }\n"
                 "td.title { background: #eee; }\n"
                 "tr.odd td { background: #ddd; }\n"
                 "tr.odd td.title { background: #bbb; }\n"
                 "</style>\n"
				 "</head>\n");
	
	BUFFER_APPEND_STRING_CONST(b, 
                 " </head>\n"
                 " <body>\n\n");
	
	
	
	BUFFER_APPEND_STRING_CONST(b, "<h1>Bandwidth Usage</h1>\n\n");
	
	BUFFER_APPEND_STRING_CONST(b, "<table cellspacing='0'>\n"
	                              "<tr><td colspan='2' class='head'>Statistics</td></tr>\n"
	                              "<tr><td class='title'>Requests:</td><td>" );
	buffer_append_long(b,*p->conf.requests_ptr);
	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n" );
	
	BUFFER_APPEND_STRING_CONST(b, "<tr class='odd'><td class='title'>KBytes Sent:</td><td>" );
	buffer_append_long(b,*p->conf.kbytes_written_ptr);
	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n" );
	
	BUFFER_APPEND_STRING_CONST(b, "<tr><td class='title'>KBytes per Request:</td><td>" );
	/* Watch out for divide by zero errors! */
	if( *p->conf.requests_ptr > 0 ) {
	    buffer_append_long(b,*p->conf.kbytes_written_ptr / *p->conf.requests_ptr);
	} else {
	    buffer_append_long(b,0);
	}
	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n" );
	
	BUFFER_APPEND_STRING_CONST(b, "<tr class='odd'><td class='title'>Transfer Limit:</td><td>" );
	
	/* Make the bytes limit easier to understand */
	if(p->conf.limit % 1048576 == 0) {
	    buffer_append_long(b,p->conf.limit/1048576);
	    BUFFER_APPEND_STRING_CONST(b, " GB");
    } else if(p->conf.limit % 1024 == 0) {
	    buffer_append_long(b,p->conf.limit/1024);
	    BUFFER_APPEND_STRING_CONST(b, " MB");
    } else {
	    buffer_append_long(b,p->conf.limit);
	    BUFFER_APPEND_STRING_CONST(b, " KB");
	}
	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n" );
	
	BUFFER_APPEND_STRING_CONST(b, "<tr><td class='title'>Limit Period:</td><td>" );

    /* Also make the time period easier to understand */
	ts = p->conf.period;
	
	days = ts / (60 * 60 * 24);
	ts %= (60 * 60 * 24);
	
	hours = ts / (60 * 60);
	ts %= (60 * 60);
	
	mins = ts / (60);
	ts %= (60);
	
	seconds = ts;
	
	if (days) {
		buffer_append_long(b, days);
		BUFFER_APPEND_STRING_CONST(b, " days ");
	}
	
	if (hours) {
		buffer_append_long(b, hours);
		BUFFER_APPEND_STRING_CONST(b, " hours ");
	}
	
	if (mins) {
		buffer_append_long(b, mins);
		BUFFER_APPEND_STRING_CONST(b, " min ");
	}
	
	if(seconds) {
    	buffer_append_long(b, seconds);
    	BUFFER_APPEND_STRING_CONST(b, " s");
    }
	
	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n"
	                              "</table>\n" );
	
	
    BUFFER_APPEND_STRING_CONST(b,"<div class='foot'>");
				 
    if (buffer_is_empty(con->conf.server_tag)) {
        BUFFER_APPEND_STRING_CONST(b, PACKAGE_NAME "/" PACKAGE_VERSION);
    } else { 
        buffer_append_string_buffer(b, con->conf.server_tag); 
    }

    BUFFER_APPEND_STRING_CONST(b,
                 "</div>\n"
                 "</body>\n"
                 "</html>\n");
	
	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
	
	con->http_status = 200;
	con->file_finished = 1;
	
	return HANDLER_FINISHED;
}

static handler_t mod_bandwidth_handle_exceeded(server *srv, connection *con, void *p_d) {
	buffer *b;
	
	UNUSED(p_d);
	
	b = chunkqueue_get_append_buffer(con->write_queue);

	BUFFER_COPY_STRING_CONST(b, 
				 "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
				 "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
				 "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
				 "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
				 "<head>\n");

    BUFFER_APPEND_STRING_CONST(b,
				 "<title>509 Bandwidth Limit Exceeded</title>\n"
				 "<style type='text/css'>\n"
				 "body{ font-family: Arial, sans-serif; }\n"
                 "div.foot { \n"
                 "    font-family: 'Courier New', Courier, monospace; font-size: 10pt; color: #787878;\n"
                 "    border-top: 1px solid #646464; margin-top: 10px; padding-top: 5px;\n}\n"
                 "</style>\n"
				 "</head>\n");
				 
    BUFFER_APPEND_STRING_CONST(b,
				 "<body>\n"
				 "<h1>Bandwidth Limit Exceeded</h1>\n"
				 "<div>The server is temporarily unable to service your request due to the site "
				 "owner reaching their bandwidth limit. Please try again later.</div>"
				 "<div class='foot'>");
				 
    if (buffer_is_empty(con->conf.server_tag)) {
        BUFFER_APPEND_STRING_CONST(b, PACKAGE_NAME "/" PACKAGE_VERSION);
    } else { 
        buffer_append_string_buffer(b, con->conf.server_tag); 
    }

    BUFFER_APPEND_STRING_CONST(b,
				 "</div>\n"
				 "</body>\n"
				 "</html>\n");
				 
	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));

    con->http_status = 200;
    con->file_finished = 1;
	
	return HANDLER_FINISHED;
}

#define PATCH(x) \
	p->conf.x = s->x;
static int mod_bandwidth_patch_connection(server *srv, connection *con, plugin_data *p) {
	size_t i, j;
	char * mult = NULL;
	plugin_config *s = p->config_storage[0];
	
	PATCH(status_url);
	PATCH(limit_buf);
	PATCH(period_buf);
	
	p->conf.start_ts_ptr = &(s->start_ts);
	p->conf.kbytes_written_ptr = &(s->kbytes_written);
	p->conf.written_overflow_ptr = &(s->written_overflow);
	p->conf.kbytes_read_ptr = &(s->kbytes_read);
	p->conf.requests_ptr = &(s->requests);
	
	/* skip the first, the global context */
	for (i = 1; i < srv->config_context->used; i++) {
		data_config *dc = (data_config *)srv->config_context->data[i];
		s = p->config_storage[i];
		
		/* condition didn't match */
		if (!config_check_cond(srv, con, dc)) continue;

        /* grab the counters */
        p->conf.start_ts_ptr = &(s->start_ts);
        p->conf.kbytes_written_ptr = &(s->kbytes_written);
	    p->conf.written_overflow_ptr = &(s->written_overflow);
        p->conf.kbytes_read_ptr = &(s->kbytes_read);
        p->conf.requests_ptr = &(s->requests);
		
		/* merge config */
		for (j = 0; j < dc->value->used; j++) {
			data_unset *du = dc->value->data[j];

			if (buffer_is_equal_string(du->key, CONST_STR_LEN("bw.limit"))) {
				PATCH(limit_buf);
			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("bw.period"))) {
				PATCH(period_buf);
			}
		}
	}
	
	/* Convert the strings back into ints */
	if(buffer_is_empty(p->conf.limit_buf)) {
	   p->conf.limit = 0;
	} else {
	   p->conf.limit  = strtol(p->conf.limit_buf->ptr, (char **) &mult, 10);
	   
		switch (*mult) {
		default:
		    break;
		case 'g':
			p->conf.limit *= 1024;
		case 'm':
			p->conf.limit *= 1024;
		case 'k': case '\0':
			break;
		}
    }

	if(buffer_is_empty(p->conf.period_buf)) {
	   p->conf.period = 0;
	} else {
	   p->conf.period  = strtol(p->conf.period_buf->ptr, (char **) &mult, 10);
	   
		switch (*mult) {
		default:
		    break;
		case 'w':
			p->conf.period *= 7;
		case 'd':
			p->conf.period *= 24;
		case 'h':
			p->conf.period *= 60;
		case 'm':
			p->conf.period *= 60;
		case 's': case '\0':
			break;
		}
    }
	
	return 0;
}

static handler_t mod_bandwidth_handler(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;

	mod_bandwidth_patch_connection(srv, con, p);

    /* Reset the limits if we've gone past the period */
    if(p->conf.period > 0 &&
       srv->cur_ts > (*(p->conf.start_ts_ptr) + (p->conf.period))) {
        *(p->conf.start_ts_ptr)         = srv->cur_ts;
        *(p->conf.requests_ptr)         = 0;
        *(p->conf.kbytes_written_ptr)   = 0;
	    *(p->conf.written_overflow_ptr) = 0;
        *(p->conf.kbytes_read_ptr)      = 0;
    }

	/* Check if we're viewing the bandwidth url */
	if (!buffer_is_empty(p->conf.status_url) && 
	    buffer_is_equal(p->conf.status_url, con->uri.path)) {
		return mod_bandwidth_handle_status(srv, con, p_d);
	/* Check if we're over the limit */
	} else if( p->conf.limit > 0 &&
	           p->conf.limit < *p->conf.kbytes_written_ptr ) {
	   	return mod_bandwidth_handle_exceeded(srv, con, p_d);
	}
	
	return HANDLER_GO_ON;
}

TRIGGER_FUNC(mod_bandwidth_trigger) {
    FILE *fp;
    size_t i;
	plugin_data *p = p_d;
	plugin_config *s = p->config_storage[0];
	
	/* Only save once every 30 to save some power */
	if ((srv->cur_ts % 5) != 0) return HANDLER_GO_ON;

    /* Open the state file */
	if ((fp = fopen(s->state_file->ptr, "w")) == NULL) {
	    log_error_write(srv, __FILE__, __LINE__, "sb"
	                    "Cannot open state file:", s->state_file);
	    return HANDLER_ERROR;
	}
	
	/* Write the state for each context */
    for (i = 0; i < srv->config_context->used; i++) {
		data_unset *du = srv->config_context->data[i];
		s = p->config_storage[i];
	
	    fprintf(fp,"%s\t%ld\t%ld\t%ld\t%ld\n",
	            du->key->ptr,
	            (long)s->start_ts,
	            (long)s->kbytes_written,
	            (long)s->written_overflow,
	            (long)s->requests );
	}

	fclose(fp);
	
	return HANDLER_GO_ON;
}

REQUESTDONE_FUNC(mod_bandwidth_account) {
	plugin_data *p = p_d;
	
	mod_bandwidth_patch_connection(srv, con, p);

	/* Skip if it's the bandwidth URL */
	if (!buffer_is_empty(p->conf.status_url) && 
	    buffer_is_equal(p->conf.status_url, con->uri.path)) {
		return HANDLER_GO_ON;
	}

    /* Add the bandwidth used */
	*(p->conf.requests_ptr)         += 1;
	*(p->conf.kbytes_written_ptr)   += con->bytes_written / 1024;
	*(p->conf.written_overflow_ptr) += con->bytes_written % 1024;
	if( *(p->conf.written_overflow_ptr) >= 1024 ) {
	    *(p->conf.kbytes_written_ptr)    += *(p->conf.written_overflow_ptr) / 1024;
	    *(p->conf.written_overflow_ptr)  = *(p->conf.written_overflow_ptr) % 1024;
	}
	*(p->conf.kbytes_read_ptr)      += con->bytes_read / 1024;
	
	return HANDLER_GO_ON;
}

int mod_bandwidth_plugin_init(plugin *p) {
	p->version     = LIGHTTPD_VERSION_ID;
	p->name        = buffer_init_string("bandwidth");
	
	p->init        = mod_bandwidth_init;
	p->cleanup     = mod_bandwidth_free;
	p->set_defaults= mod_bandwidth_set_defaults;
	
	p->handle_uri_clean    = mod_bandwidth_handler;
	p->handle_trigger      = mod_bandwidth_trigger;
	p->handle_request_done = mod_bandwidth_account;
	
	p->data        = NULL;
	
	return 0;
}
